@@ -1,16 +1,23 @@
 import collections.abc
 import datetime
 import gzip
+import io
 import json
-import re
+import zlib
 from concurrent.futures import ThreadPoolExecutor
+from typing import AsyncIterator, Optional
 from unittest import mock
 
+import aiosignal
 import pytest
 from multidict import CIMultiDict, CIMultiDictProxy
+from re_assert import Matches
 
-from aiohttp import HttpVersion, HttpVersion10, HttpVersion11, hdrs, signals
-from aiohttp.payload import BytesPayload
+from aiohttp import HttpVersion, HttpVersion10, HttpVersion11, hdrs
+from aiohttp.helpers import ETag
+from aiohttp.http_writer import StreamWriter, _serialize_headers
+from aiohttp.multipart import BodyPartReader, MultipartWriter
+from aiohttp.payload import BytesPayload, StringPayload
 from aiohttp.test_utils import make_mocked_coro, make_mocked_request
 from aiohttp.web import ContentCoding, Response, StreamResponse, json_response
 
@@ -21,12 +28,12 @@ def make_request(
     headers=CIMultiDict(),
     version=HttpVersion11,
     on_response_prepare=None,
-    **kwargs
+    **kwargs,
 ):
     app = kwargs.pop("app", None) or mock.Mock()
     app._debug = False
     if on_response_prepare is None:
-        on_response_prepare = signals.Signal(app)
+        on_response_prepare = aiosignal.Signal(app)
     app.on_response_prepare = on_response_prepare
     app.on_response_prepare.freeze()
     protocol = kwargs.pop("protocol", None) or mock.Mock()
@@ -54,12 +61,7 @@ def writer(buf):
         buf.extend(chunk)
 
     async def write_headers(status_line, headers):
-        headers = (
-            status_line
-            + "\r\n"
-            + "".join([k + ": " + v + "\r\n" for k, v in headers.items()])
-        )
-        headers = headers.encode("utf-8") + b"\r\n"
+        headers = _serialize_headers(status_line, headers)
         buf.extend(headers)
 
     async def write_eof(chunk=b""):
@@ -253,6 +255,95 @@ def test_last_modified_reset() -> None:
     assert resp.last_modified is None
 
 
+@pytest.mark.parametrize(
+    ["header_val", "expected"],
+    [
+        pytest.param("xxyyzz", None),
+        pytest.param("Tue, 08 Oct 4446413 00:56:40 GMT", None),
+        pytest.param("Tue, 08 Oct 2000 00:56:80 GMT", None),
+    ],
+)
+def test_last_modified_string_invalid(header_val, expected) -> None:
+    resp = StreamResponse(headers={"Last-Modified": header_val})
+    assert resp.last_modified == expected
+
+
+def test_etag_initial() -> None:
+    resp = StreamResponse()
+    assert resp.etag is None
+
+
+def test_etag_string() -> None:
+    resp = StreamResponse()
+    value = "0123-kotik"
+    resp.etag = value
+    assert resp.etag == ETag(value=value)
+    assert resp.headers[hdrs.ETAG] == f'"{value}"'
+
+
+@pytest.mark.parametrize(
+    ["etag", "expected_header"],
+    (
+        (ETag(value="0123-weak-kotik", is_weak=True), 'W/"0123-weak-kotik"'),
+        (ETag(value="0123-strong-kotik", is_weak=False), '"0123-strong-kotik"'),
+    ),
+)
+def test_etag_class(etag, expected_header) -> None:
+    resp = StreamResponse()
+    resp.etag = etag
+    assert resp.etag == etag
+    assert resp.headers[hdrs.ETAG] == expected_header
+
+
+def test_etag_any() -> None:
+    resp = StreamResponse()
+    resp.etag = "*"
+    assert resp.etag == ETag(value="*")
+    assert resp.headers[hdrs.ETAG] == "*"
+
+
+@pytest.mark.parametrize(
+    "invalid_value",
+    (
+        '"invalid"',
+        "повинен бути ascii",
+        ETag(value='"invalid"', is_weak=True),
+        ETag(value="bad ©®"),
+    ),
+)
+def test_etag_invalid_value_set(invalid_value) -> None:
+    resp = StreamResponse()
+    with pytest.raises(ValueError, match="is not a valid etag"):
+        resp.etag = invalid_value
+
+
+@pytest.mark.parametrize(
+    "header",
+    (
+        "forgotten quotes",
+        '"∀ x ∉ ascii"',
+    ),
+)
+def test_etag_invalid_value_get(header) -> None:
+    resp = StreamResponse()
+    resp.headers["ETag"] = header
+    assert resp.etag is None
+
+
+@pytest.mark.parametrize("invalid", (123, ETag(value=123, is_weak=True)))
+def test_etag_invalid_value_class(invalid) -> None:
+    resp = StreamResponse()
+    with pytest.raises(ValueError, match="Unsupported etag type"):
+        resp.etag = invalid
+
+
+def test_etag_reset() -> None:
+    resp = StreamResponse()
+    resp.etag = "*"
+    resp.etag = None
+    assert resp.etag is None
+
+
 async def test_start() -> None:
     req = make_request("GET", "/")
     resp = StreamResponse()
@@ -314,7 +405,7 @@ async def test_chunked_encoding_forbidden_for_http_10() -> None:
 
     with pytest.raises(RuntimeError) as ctx:
         await resp.prepare(req)
-    assert re.match("Using chunked encoding is forbidden for HTTP/1.0", str(ctx.value))
+    assert Matches("Using chunked encoding is forbidden for HTTP/1.0") == str(ctx.value)
 
 
 async def test_compression_no_accept() -> None:
@@ -371,7 +462,7 @@ async def test_compression_default_coding() -> None:
 
     msg = await resp.prepare(req)
 
-    msg.enable_compression.assert_called_with("deflate")
+    msg.enable_compression.assert_called_with("deflate", zlib.Z_DEFAULT_STRATEGY)
     assert "deflate" == resp.headers.get(hdrs.CONTENT_ENCODING)
     assert msg.filter is not None
 
@@ -386,7 +477,25 @@ async def test_force_compression_deflate() -> None:
     assert resp.compression
 
     msg = await resp.prepare(req)
-    msg.enable_compression.assert_called_with("deflate")
+    msg.enable_compression.assert_called_with("deflate", zlib.Z_DEFAULT_STRATEGY)
+    assert "deflate" == resp.headers.get(hdrs.CONTENT_ENCODING)
+
+
+async def test_force_compression_deflate_large_payload() -> None:
+    """Make sure a warning is thrown for large payloads compressed in the event loop."""
+    req = make_request(
+        "GET", "/", headers=CIMultiDict({hdrs.ACCEPT_ENCODING: "gzip, deflate"})
+    )
+    resp = Response(body=b"large")
+
+    resp.enable_compression(ContentCoding.deflate)
+    assert resp.compression
+
+    with pytest.warns(
+        Warning, match="Synchronous compression of large response bodies"
+    ), mock.patch("aiohttp.web_response.LARGE_BODY_SIZE", 2):
+        msg = await resp.prepare(req)
+        assert msg is not None
     assert "deflate" == resp.headers.get(hdrs.CONTENT_ENCODING)
 
 
@@ -398,7 +507,7 @@ async def test_force_compression_no_accept_deflate() -> None:
     assert resp.compression
 
     msg = await resp.prepare(req)
-    msg.enable_compression.assert_called_with("deflate")
+    msg.enable_compression.assert_called_with("deflate", zlib.Z_DEFAULT_STRATEGY)
     assert "deflate" == resp.headers.get(hdrs.CONTENT_ENCODING)
 
 
@@ -412,7 +521,7 @@ async def test_force_compression_gzip() -> None:
     assert resp.compression
 
     msg = await resp.prepare(req)
-    msg.enable_compression.assert_called_with("gzip")
+    msg.enable_compression.assert_called_with("gzip", zlib.Z_DEFAULT_STRATEGY)
     assert "gzip" == resp.headers.get(hdrs.CONTENT_ENCODING)
 
 
@@ -424,7 +533,7 @@ async def test_force_compression_no_accept_gzip() -> None:
     assert resp.compression
 
     msg = await resp.prepare(req)
-    msg.enable_compression.assert_called_with("gzip")
+    msg.enable_compression.assert_called_with("gzip", zlib.Z_DEFAULT_STRATEGY)
     assert "gzip" == resp.headers.get(hdrs.CONTENT_ENCODING)
 
 
@@ -555,7 +664,6 @@ async def test_rm_content_length_if_compression_http11() -> None:
     req = make_request("GET", "/", writer=writer)
     payload = BytesPayload(b"answer", headers={"X-Test-Header": "test"})
     resp = Response(body=payload)
-    assert resp.content_length == 6
     resp.body = payload
     resp.enable_compression(ContentCoding.gzip)
     await resp.prepare(req)
@@ -577,6 +685,90 @@ async def test_rm_content_length_if_compression_http10() -> None:
     assert resp.content_length is None
 
 
+async def test_rm_content_length_if_204() -> None:
+    """Ensure content-length is removed for 204 responses."""
+    writer = mock.create_autospec(StreamWriter, spec_set=True, instance=True)
+
+    async def write_headers(status_line, headers):
+        assert hdrs.CONTENT_LENGTH not in headers
+
+    writer.write_headers.side_effect = write_headers
+    req = make_request("GET", "/", writer=writer)
+    payload = BytesPayload(b"answer", headers={"Content-Length": "6"})
+    resp = Response(body=payload, status=204)
+    resp.body = payload
+    await resp.prepare(req)
+    assert resp.content_length is None
+
+
+@pytest.mark.parametrize("status", (100, 101, 204, 304))
+async def test_rm_transfer_encoding_rfc_9112_6_3_http_11(status: int) -> None:
+    """Remove transfer encoding for RFC 9112 sec 6.3 with HTTP/1.1."""
+    writer = mock.create_autospec(StreamWriter, spec_set=True, instance=True)
+    req = make_request("GET", "/", version=HttpVersion11, writer=writer)
+    resp = Response(status=status, headers={hdrs.TRANSFER_ENCODING: "chunked"})
+    await resp.prepare(req)
+    assert resp.content_length == 0
+    assert not resp.chunked
+    assert hdrs.CONTENT_LENGTH not in resp.headers
+    assert hdrs.TRANSFER_ENCODING not in resp.headers
+
+
+@pytest.mark.parametrize("status", (100, 101, 102, 204, 304))
+async def test_rm_content_length_1xx_204_304_responses(status: int) -> None:
+    """Remove content length for 1xx, 204, and 304 responses.
+
+    Content-Length is forbidden for 1xx and 204
+    https://datatracker.ietf.org/doc/html/rfc7230#section-3.3.2
+
+    Content-Length is discouraged for 304.
+    https://datatracker.ietf.org/doc/html/rfc7232#section-4.1
+    """
+    writer = mock.create_autospec(StreamWriter, spec_set=True, instance=True)
+    req = make_request("GET", "/", version=HttpVersion11, writer=writer)
+    resp = Response(status=status, body="answer")
+    await resp.prepare(req)
+    assert not resp.chunked
+    assert hdrs.CONTENT_LENGTH not in resp.headers
+    assert hdrs.TRANSFER_ENCODING not in resp.headers
+
+
+async def test_head_response_keeps_content_length_of_original_body() -> None:
+    """Verify HEAD response keeps the content length of the original body HTTP/1.1."""
+    writer = mock.create_autospec(StreamWriter, spec_set=True, instance=True)
+    req = make_request("HEAD", "/", version=HttpVersion11, writer=writer)
+    resp = Response(status=200, body=b"answer")
+    await resp.prepare(req)
+    assert resp.content_length == 6
+    assert not resp.chunked
+    assert resp.headers[hdrs.CONTENT_LENGTH] == "6"
+    assert hdrs.TRANSFER_ENCODING not in resp.headers
+
+
+async def test_head_response_omits_content_length_when_body_unset() -> None:
+    """Verify HEAD response omits content-length body when its unset."""
+    writer = mock.create_autospec(StreamWriter, spec_set=True, instance=True)
+    req = make_request("HEAD", "/", version=HttpVersion11, writer=writer)
+    resp = Response(status=200)
+    await resp.prepare(req)
+    assert resp.content_length == 0
+    assert not resp.chunked
+    assert hdrs.CONTENT_LENGTH not in resp.headers
+    assert hdrs.TRANSFER_ENCODING not in resp.headers
+
+
+async def test_304_response_omits_content_length_when_body_unset() -> None:
+    """Verify 304 response omits content-length body when its unset."""
+    writer = mock.create_autospec(StreamWriter, spec_set=True, instance=True)
+    req = make_request("GET", "/", version=HttpVersion11, writer=writer)
+    resp = Response(status=304)
+    await resp.prepare(req)
+    assert resp.content_length == 0
+    assert not resp.chunked
+    assert hdrs.CONTENT_LENGTH not in resp.headers
+    assert hdrs.TRANSFER_ENCODING not in resp.headers
+
+
 async def test_content_length_on_chunked() -> None:
     req = make_request("GET", "/")
     resp = Response(body=b"answer")
@@ -619,11 +811,8 @@ async def test___repr___after_eof() -> None:
     resp = StreamResponse()
     await resp.prepare(make_request("GET", "/"))
 
-    assert resp.prepared
-
     await resp.write(b"data")
     await resp.write_eof()
-    assert not resp.prepared
     resp_repr = repr(resp)
     assert resp_repr == "<StreamResponse OK eof>"
 
@@ -689,7 +878,7 @@ def test_response_cookies() -> None:
         'Set-Cookie: name=("")?; '
         "expires=Thu, 01 Jan 1970 00:00:00 GMT; Max-Age=0; Path=/"
     )
-    assert re.match(expected, str(resp.cookies))
+    assert Matches(expected) == str(resp.cookies)
 
     resp.set_cookie("name", "value", domain="local.host")
     expected = "Set-Cookie: name=value; Domain=local.host; Path=/"
@@ -741,7 +930,7 @@ def test_response_cookie__issue_del_cookie() -> None:
         'Set-Cookie: name=("")?; '
         "expires=Thu, 01 Jan 1970 00:00:00 GMT; Max-Age=0; Path=/"
     )
-    assert re.match(expected, str(resp.cookies))
+    assert Matches(expected) == str(resp.cookies)
 
 
 def test_cookie_set_after_del() -> None:
@@ -757,9 +946,17 @@ def test_cookie_set_after_del() -> None:
 def test_set_status_with_reason() -> None:
     resp = StreamResponse()
 
-    resp.set_status(200, "Everithing is fine!")
+    resp.set_status(200, "Everything is fine!")
     assert 200 == resp.status
-    assert "Everithing is fine!" == resp.reason
+    assert "Everything is fine!" == resp.reason
+
+
+def test_set_status_with_empty_reason() -> None:
+    resp = StreamResponse()
+
+    resp.set_status(200, "")
+    assert resp.status == 200
+    assert resp.reason == ""
 
 
 async def test_start_force_close() -> None:
@@ -774,14 +971,14 @@ async def test_start_force_close() -> None:
 
 async def test___repr__() -> None:
     req = make_request("GET", "/path/to")
-    resp = StreamResponse(reason=301)
+    resp = StreamResponse(reason="foo")
     await resp.prepare(req)
-    assert "<StreamResponse 301 GET /path/to >" == repr(resp)
+    assert "<StreamResponse foo GET /path/to >" == repr(resp)
 
 
 def test___repr___not_prepared() -> None:
-    resp = StreamResponse(reason=301)
-    assert "<StreamResponse 301 not prepared>" == repr(resp)
+    resp = StreamResponse(reason="foo")
+    assert "<StreamResponse foo not prepared>" == repr(resp)
 
 
 async def test_keep_alive_http10_default() -> None:
@@ -820,7 +1017,7 @@ async def test_prepare_twice() -> None:
 async def test_prepare_calls_signal() -> None:
     app = mock.Mock()
     sig = make_mocked_coro()
-    on_response_prepare = signals.Signal(app)
+    on_response_prepare = aiosignal.Signal(app)
     on_response_prepare.append(sig)
     req = make_request("GET", "/", app=app, on_response_prepare=on_response_prepare)
     resp = StreamResponse()
@@ -968,6 +1165,48 @@ def test_assign_nonstr_text() -> None:
     assert 4 == resp.content_length
 
 
+mpwriter = MultipartWriter(boundary="x")
+mpwriter.append_payload(StringPayload("test"))
+
+
+async def async_iter() -> AsyncIterator[str]:
+    yield "foo"  # pragma: no cover
+
+
+class CustomIO(io.IOBase):
+    def __init__(self):
+        self._lines = [b"", b"", b"test"]
+
+    def read(self, size: int = -1) -> bytes:
+        return self._lines.pop()
+
+
+@pytest.mark.parametrize(
+    "payload,expected",
+    (
+        ("test", "test"),
+        (CustomIO(), "test"),
+        (io.StringIO("test"), "test"),
+        (io.TextIOWrapper(io.BytesIO(b"test")), "test"),
+        (io.BytesIO(b"test"), "test"),
+        (io.BufferedReader(io.BytesIO(b"test")), "test"),
+        (async_iter(), None),
+        (BodyPartReader("x", CIMultiDictProxy(CIMultiDict()), mock.Mock()), None),
+        (
+            mpwriter,
+            "--x\nContent-Type: text/plain; charset=utf-8\r\nContent-Length: 4\r\n\r\ntest",
+        ),
+    ),
+)
+def test_payload_body_get_text(payload, expected: Optional[str]) -> None:
+    resp = Response(body=payload)
+    if expected is None:
+        with pytest.raises(TypeError):
+            resp.text
+    else:
+        assert resp.text == expected
+
+
 def test_response_set_content_length() -> None:
     resp = Response()
     with pytest.raises(RuntimeError):
@@ -981,13 +1220,14 @@ async def test_send_headers_for_empty_body(buf, writer) -> None:
     await resp.prepare(req)
     await resp.write_eof()
     txt = buf.decode("utf8")
-    assert re.match(
-        "HTTP/1.1 200 OK\r\n"
-        "Content-Length: 0\r\n"
-        "Content-Type: application/octet-stream\r\n"
-        "Date: .+\r\n"
-        "Server: .+\r\n\r\n",
-        txt,
+    assert (
+        Matches(
+            "HTTP/1.1 200 OK\r\n"
+            "Content-Length: 0\r\n"
+            "Date: .+\r\n"
+            "Server: .+\r\n\r\n"
+        )
+        == txt
     )
 
 
@@ -999,17 +1239,24 @@ async def test_render_with_body(buf, writer) -> None:
     await resp.write_eof()
 
     txt = buf.decode("utf8")
-    assert re.match(
-        "HTTP/1.1 200 OK\r\n"
-        "Content-Length: 4\r\n"
-        "Content-Type: application/octet-stream\r\n"
-        "Date: .+\r\n"
-        "Server: .+\r\n\r\n"
-        "data",
-        txt,
+    assert (
+        Matches(
+            "HTTP/1.1 200 OK\r\n"
+            "Content-Length: 4\r\n"
+            "Content-Type: application/octet-stream\r\n"
+            "Date: .+\r\n"
+            "Server: .+\r\n\r\n"
+            "data"
+        )
+        == txt
     )
 
 
+async def test_multiline_reason(buf, writer) -> None:
+    with pytest.raises(ValueError, match=r"Reason cannot contain \\n"):
+        Response(reason="Bad\r\nInjected-header: foo")
+
+
 async def test_send_set_cookie_header(buf, writer) -> None:
     resp = Response()
     resp.cookies["name"] = "value"
@@ -1019,14 +1266,15 @@ async def test_send_set_cookie_header(buf, writer) -> None:
     await resp.write_eof()
 
     txt = buf.decode("utf8")
-    assert re.match(
-        "HTTP/1.1 200 OK\r\n"
-        "Content-Length: 0\r\n"
-        "Set-Cookie: name=value\r\n"
-        "Content-Type: application/octet-stream\r\n"
-        "Date: .+\r\n"
-        "Server: .+\r\n\r\n",
-        txt,
+    assert (
+        Matches(
+            "HTTP/1.1 200 OK\r\n"
+            "Content-Length: 0\r\n"
+            "Set-Cookie: name=value\r\n"
+            "Date: .+\r\n"
+            "Server: .+\r\n\r\n"
+        )
+        == txt
     )
 
 
@@ -1085,17 +1333,25 @@ def test_content_type_with_set_body() -> None:
     assert resp.content_type == "application/octet-stream"
 
 
-def test_started_when_not_started() -> None:
+def test_prepared_when_not_started() -> None:
     resp = StreamResponse()
     assert not resp.prepared
 
 
-async def test_started_when_started() -> None:
+async def test_prepared_when_started() -> None:
     resp = StreamResponse()
     await resp.prepare(make_request("GET", "/"))
     assert resp.prepared
 
 
+async def test_prepared_after_eof() -> None:
+    resp = StreamResponse()
+    await resp.prepare(make_request("GET", "/"))
+    await resp.write(b"data")
+    await resp.write_eof()
+    assert resp.prepared
+
+
 async def test_drain_before_start() -> None:
     resp = StreamResponse()
     with pytest.raises(AssertionError):
@@ -1178,7 +1434,7 @@ async def test_response_prepared_after_header_preparation() -> None:
             del res.headers["Server"]
 
     app = mock.Mock()
-    sig = signals.Signal(app)
+    sig = aiosignal.Signal(app)
     sig.append(_strip_server)
 
     req = make_request("GET", "/", on_response_prepare=sig, app=app)
